@@ -5,11 +5,11 @@ import json
 import os
 import re
 import types
-from typing import Optional, Union
+from typing import List, Optional, Union
 
 import requests
-from bs4 import BeautifulSoup
 
+from seeq.base import util
 from seeq.base.seeq_names import SeeqNames
 from seeq.sdk import *
 from seeq.spy import _common
@@ -22,6 +22,7 @@ from seeq.spy.workbooks import _item, _render
 from seeq.spy.workbooks._content import Content, DateRange, AssetSelection
 from seeq.spy.workbooks._item import Item
 from seeq.spy.workbooks._item_map import ItemMap, OverrideItemMap
+from seeq.spy.workbooks._user import ItemWithOwnerAndAcl, Identity
 from seeq.spy.workbooks._workstep import AnalysisWorkstep
 
 FIXUP_PREFIX = 'Annotation to fix up'
@@ -51,7 +52,9 @@ class Annotation(Item):
     def id(self, _id):
         self.worksheet['Annotation ID'] = _id
 
-    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
+    def refresh_from(self, new_item, item_map: ItemMap, status: Status, session: Session):
+        super().refresh_from(new_item, item_map, status, session)
+
         self.annotation_type = new_item.annotation_type
         # Note that we purposefully don't touch the worksheet reference, since it will have stayed the same
         self._set_html(new_item.html)
@@ -194,7 +197,12 @@ class Annotation(Item):
         return annotation_output
 
     def push(self, session: Session, pushed_workbook_id, pushed_worksheet_id, item_map, datasource_output,
-             push_images, label, status=None):
+             access_control, push_images, label, status=None):
+        try:
+            from bs4 import BeautifulSoup
+        except ImportError:
+            raise SPyDependencyNotFound(f'`BeautifulSoup4` is required to push annotations. Please install it using '
+                                        f'`pip install seeq-spy[templates]` to use this feature.')
         self.render_plots()
 
         new_annotation = AnnotationInputV1()
@@ -233,7 +241,8 @@ class Annotation(Item):
 
         item_map[self.id] = relevant_annotation.id
 
-        html = self._push_specific(session, item_map, datasource_output, label, new_annotation, relevant_annotation)
+        html = self._push_specific(session, item_map, datasource_output, label, new_annotation, relevant_annotation,
+                                   access_control, status=status)
 
         # Create an override map for image mapping, which can be different for the same annotation in the case of a
         # template
@@ -307,6 +316,7 @@ class Annotation(Item):
         new_annotation.description = description
         new_annotation.document = html
         new_annotation.interests = list()
+
         for interest in relevant_annotation.interests:  # type: AnnotationInterestOutputV1
             interest_item = interest.item  # type: ItemPreviewV1
             # At Chevron, we encountered a case where there were multiple interests returned with the same ID, which
@@ -329,8 +339,18 @@ class Annotation(Item):
 
         # Annotations have a lot of properties that are managed by Appserver and shouldn't be pushed. We need to push
         # only a small, curated set that are actually configuration items on the Topic Document.
-        properties_to_push = [SeeqNames.Properties.timezone,
-                              SeeqNames.Properties.is_ck_enabled]
+        properties_to_push = [
+            SeeqNames.Properties.margin_top,
+            SeeqNames.Properties.margin_bottom,
+            SeeqNames.Properties.margin_left,
+            SeeqNames.Properties.margin_right,
+            SeeqNames.Properties.orientation,
+            SeeqNames.Properties.page_size,
+            SeeqNames.Properties.is_fixed_with,
+            SeeqNames.Properties.discoverable,
+            SeeqNames.Properties.timezone,
+            SeeqNames.Properties.is_ck_enabled
+        ]
 
         added_or_updated_properties = [ScalarPropertyV1(name=k, value=v)
                                        for k, v in self.definition_dict.items()
@@ -385,8 +405,9 @@ class Annotation(Item):
 
         return html
 
-    def _push_specific(self, session: Session, item_map, datasource_output, label, new_annotation, annotation,
-                       override_content_dict=None, status=None):
+    def _push_specific(self, session: Session, item_map, datasource_output, label,
+                       new_annotation: AnnotationInputV1, annotation: AnnotationOutputV1,
+                       access_control, override_content_dict=None, status=None):
         # This will be overridden in derived classes to do some specific work like pushing Content and DateRanges
         return self.html
 
@@ -405,9 +426,15 @@ class Annotation(Item):
     def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
         self.render_plots()
         html_file = f'{self._get_annotation_file(workbook_folder)}.html'
-        with open(html_file, 'w', encoding='utf-8') as f:
+        with util.safe_open(html_file, 'w', encoding='utf-8') as f:
             if self.html:
                 if pretty_print_html:
+                    try:
+                        from bs4 import BeautifulSoup
+                    except ImportError:
+                        raise SPyDependencyNotFound(f'`BeautifulSoup4` is required to save pretty-printed HTML. Please '
+                                                    f'install it using `pip install seeq-spy[templates]`'
+                                                    f' to use this feature.')
                     html_to_save = BeautifulSoup(self.html, features='html.parser').prettify()
                     html_to_save = _common.fix_up_ckeditor_curly_brace_weirdness(html_to_save)
                     # If we don't trim the spaces within <a> tags, you'll get extra spaces underlined in the UI
@@ -424,7 +451,7 @@ class Annotation(Item):
             'Definition': self.definition_dict,
         }
         self._update_save_json_dict(json_dict)
-        with open(json_file, 'w', encoding='utf-8') as f:
+        with util.safe_open(json_file, 'w', encoding='utf-8') as f:
             json.dump(json_dict, f, indent=4, sort_keys=True)
 
     def _update_save_json_dict(self, json_dict):
@@ -435,8 +462,8 @@ class Annotation(Item):
         html_file = f'{self._get_annotation_file(workbook_folder)}.html'
         json_file = f'{self._get_annotation_file(workbook_folder)}.json'
 
-        if os.path.exists(json_file):
-            with open(json_file, 'r', encoding='utf-8') as f:
+        if util.safe_exists(json_file):
+            with util.safe_open(json_file, 'r', encoding='utf-8') as f:
                 json_dict = json.load(f)
         else:
             json_dict = dict()
@@ -444,7 +471,7 @@ class Annotation(Item):
         if 'Definition' in json_dict:
             self._definition = json_dict['Definition']
 
-        with open(html_file, 'r', encoding='utf-8') as f:
+        with util.safe_open(html_file, 'r', encoding='utf-8') as f:
             self.html = f.read()
 
         matches = re.finditer(r'src="/api(/annotations/(.*?)/images/(.*?))"', self.html)
@@ -452,7 +479,7 @@ class Annotation(Item):
             image_id_tuple = (match.group(2), match.group(3))
             image_file = _common.get_image_file(workbook_folder, image_id_tuple)
 
-            with open(image_file, 'rb') as f:
+            with util.safe_open(image_file, 'rb') as f:
                 self.images[image_id_tuple] = f.read()
 
         return json_dict
@@ -490,7 +517,7 @@ class Annotation(Item):
         html = self.html
         if filename:
             image_name = os.path.basename(filename)
-            with open(filename, 'rb') as img:
+            with util.safe_open(filename, 'rb') as img:
                 self.images[(self.id, image_name)] = img.read()
         else:
             image_name = f'{_common.new_placeholder_guid()}.{image_format}'
@@ -536,15 +563,15 @@ class Journal(Annotation):
 
     @property
     def referenced_items(self):
-        referenced_items = list()
+        referenced_items = set()
         if self.html:
             matches = re.finditer(r'item%s(%s)' % (_common.HTML_EQUALS_REGEX, _common.GUID_REGEX), self.html,
                                   re.IGNORECASE)
             for match in matches:
-                referenced_items.append(_item.Reference(match.group(1).upper(), _item.Reference.JOURNAL,
-                                                        self.worksheet))
+                referenced_items.add(_item.Reference(match.group(1).upper(), _item.Reference.JOURNAL,
+                                                     self.worksheet))
 
-        return referenced_items
+        return list(referenced_items)
 
     def find_workstep_references(self, *, item_map: Optional[ItemMap] = None):
         if not self.html:
@@ -579,20 +606,20 @@ class Report(Annotation):
 
     @property
     def referenced_items(self):
-        referenced_items = list()
+        referenced_items = set()
 
         for date_range in self.date_ranges.values():
             if _common.present(date_range, 'Condition ID'):
-                referenced_items.append(
+                referenced_items.add(
                     _item.Reference(date_range['Condition ID'], _item.Reference.DATE_RANGE_CONDITION,
                                     self.worksheet))
 
         for asset_selection in self.asset_selections.values():
             if _common.present(asset_selection, 'Asset ID'):
-                referenced_items.append(_item.Reference(asset_selection['Asset ID'], _item.Reference.ASSET_SELECTION,
-                                                        self.worksheet))
+                referenced_items.add(_item.Reference(asset_selection['Asset ID'], _item.Reference.ASSET_SELECTION,
+                                                     self.worksheet))
 
-        return referenced_items
+        return list(referenced_items)
 
     def pull(self, session: Session, include_images=True, status=None):
         session = Session.validate(session)
@@ -600,6 +627,25 @@ class Report(Annotation):
         if annotation_output is None:
             return
 
+        # A lot of these come down as item properties, but we map them here anyways just to be double-sure.
+        # This also converts date/times to human-readable.
+        self._definition.update(Item._dict_via_attribute_map(annotation_output, {
+            'ck_enabled': SeeqNames.Properties.is_ck_enabled,
+            'created_at': 'Created At',
+            'updated_at': 'Updated At',
+            'discoverable': SeeqNames.Properties.discoverable,
+            'fixed_width': SeeqNames.Properties.is_fixed_with,
+            'margin_bottom': SeeqNames.Properties.margin_bottom,
+            'margin_left': SeeqNames.Properties.margin_left,
+            'margin_right': SeeqNames.Properties.margin_right,
+            'margin_top': SeeqNames.Properties.margin_top,
+            'next_run_time': 'Next Run Time',
+            'page_orientation': SeeqNames.Properties.orientation,
+            'page_size': SeeqNames.Properties.page_size,
+            'published_at': 'Published At',
+            'timezone': SeeqNames.Properties.timezone,
+        }))
+
         content_api = ContentApi(session.client)
         contents = safely(
             lambda: content_api.get_contents_with_all_metadata(report_id=annotation_output.id),
@@ -641,11 +687,68 @@ class Report(Annotation):
         if annotation_output.cron_schedule:
             self.schedule = {
                 'Cron Schedule': annotation_output.cron_schedule,
-                'Background': annotation_output.background
+                'Background': annotation_output.background,
+                'Enabled': annotation_output.enabled
             }
 
+            self._pull_notification_config(status, annotation_output)
+
+    def _pull_notification_config(self, status: Status, annotation_output):
+        # noinspection PyBroadException
+        try:
+            notification_configurations_api = NotificationConfigurationsApi(status.session.client)
+            notification_config = notification_configurations_api.get_notification_configuration(
+                id=annotation_output.id)
+
+            self.schedule['Notification'] = {
+                'To Email Recipients': Report._email_recipients_to_dict(
+                    status, notification_config.to_email_recipients
+                ),
+                'CC Email Recipients': Report._email_recipients_to_dict(
+                    status, notification_config.cc_email_recipients
+                ),
+                'BCC Email Recipients': Report._email_recipients_to_dict(
+                    status, notification_config.bcc_email_recipients
+                ),
+                'Report Format': notification_config.report_format
+            }
+
+        except NameError:
+            # We're on an earlier version of Seeq that doesn't have the NotificationConfigurationsApi
+            pass
+
+        except ApiException as e:
+            # A 404 error means that the Notification Configuration doesn't exist on this document
+            if e.status != 404:
+                raise
+
     @staticmethod
-    def _refresh_child_dict(existing_item_dict: dict, new_item_dict: dict, item_map: ItemMap, status: Status):
+    def _email_recipients_to_dict(status: Status, email_recipients: List[EmailNotificationRecipientOutputV1]):
+        recipient_dicts = list()
+        for email_recipient in email_recipients:
+            recipient_dict = dict()
+            if email_recipient.identity_id is not None:
+                identity = Identity.pull(email_recipient.identity_id, status=status, session=status.session)
+                recipient_dict['Identity'] = identity.definition_dict
+                recipient_dict['Email'] = identity.get('Email')
+                recipient_dict['Name'] = identity.get('Name')
+
+            if email_recipient.email_address:
+                recipient_dict['Email'] = email_recipient.email_address
+
+            if email_recipient.name:
+                recipient_dict['Name'] = email_recipient.name
+
+            if email_recipient.redacted is not None:
+                recipient_dict['Redacted'] = email_recipient.redacted
+
+            recipient_dicts.append(recipient_dict)
+
+        return recipient_dicts
+
+    @staticmethod
+    def _refresh_child_dict(existing_item_dict: dict, new_item_dict: dict, item_map: ItemMap, status: Status,
+                            session: Session):
         items = list(existing_item_dict.items())
         for _original_id, existing_item in items:
             if existing_item.id not in item_map or item_map[existing_item.id] not in new_item_dict:
@@ -661,17 +764,20 @@ class Report(Annotation):
             existing_item_dict[new_item_id] = existing_item
 
             # Refresh item itself
-            existing_item.refresh_from(new_item, item_map, status)
+            existing_item.refresh_from(new_item, item_map, status, session)
 
-    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
-        super().refresh_from(new_item, item_map, status)
+    def refresh_from(self, new_item, item_map: ItemMap, status: Status, session: Session):
+        super().refresh_from(new_item, item_map, status, session)
 
-        Report._refresh_child_dict(self.date_ranges, new_item.date_ranges, item_map, status)
-        Report._refresh_child_dict(self.asset_selections, new_item.asset_selections, item_map, status)
-        Report._refresh_child_dict(self.content, new_item.content, item_map, status)
+        Report._refresh_child_dict(self.date_ranges, new_item.date_ranges, item_map, status, session)
+        Report._refresh_child_dict(self.asset_selections, new_item.asset_selections, item_map, status, session)
+        Report._refresh_child_dict(self.content, new_item.content, item_map, status, session)
 
-    def _push_specific(self, session: Session, item_map: ItemMap, datasource_output, label, new_annotation,
-                       existing_annotation, override_content_dict=None, status=None):
+        self.schedule = new_item.schedule
+
+    def _push_specific(self, session: Session, item_map: ItemMap, datasource_output, label,
+                       new_annotation: AnnotationInputV1, existing_annotation: AnnotationOutputV1,
+                       access_control, override_content_dict=None, status=None):
         date_range_ids_to_archive = list(existing_annotation.date_range_ids)
         asset_selection_ids_to_archive = list(existing_annotation.asset_selection_ids)
         content_ids_to_archive = list(existing_annotation.content_ids)
@@ -711,7 +817,7 @@ class Report(Annotation):
                     if _item_output.id in _ids_to_archive:
                         _ids_to_archive.remove(_item_output.id)
                 except ApiException as e:
-                    self.worksheet.workbook.status.raise_or_catalog(
+                    self.worksheet.workbook.status.on_error(
                         f'Error processing {_item_type}: {_item_object}\n{_common.format_exception(e)}')
 
         _push_it('Date Range', self.date_ranges, existing_date_ranges, date_range_ids_to_archive)
@@ -720,7 +826,7 @@ class Report(Annotation):
                  existing_contents, content_ids_to_archive)
 
         if self.schedule is not None:
-            optional_report.enabled = True
+            optional_report.enabled = self.schedule.get('Enabled', self.get('Enabled', True))
             optional_report.background = self.schedule['Background']
             optional_report.cron_schedule = self.schedule['Cron Schedule']
 
@@ -748,6 +854,8 @@ class Report(Annotation):
 
             _archive_content()
 
+        self._push_notification_config(status, item_map, access_control, existing_annotation.id)
+
         for asset_selection_id_to_archive in asset_selection_ids_to_archive:
             safely(
                 lambda: items_api.archive_item(id=asset_selection_id_to_archive),
@@ -766,6 +874,52 @@ class Report(Annotation):
 
         return self.html
 
+    def _push_notification_config(self, status: Status, item_map: ItemMap, access_control: str,
+                                  existing_annotation_id: str):
+        if self.schedule is None or self.schedule.get('Notification') is None:
+            return
+
+        _, strict = ItemWithOwnerAndAcl.parse_access_control_str(access_control)
+        datasource_maps = self.worksheet.workbook.datasource_maps
+
+        def _lookup_identity(_recipient):
+            if _recipient.get('Identity') is None:
+                return _recipient.get('Email')
+
+            try:
+                return Identity.find_identity(status.session, _recipient['Identity'], datasource_maps, item_map)
+            except SPyDependencyNotFound:
+                if strict:
+                    raise
+                else:
+                    return None
+
+        def _lookup_identities(_recipients):
+            _identities = {_lookup_identity(r) for r in _recipients}
+            return [_id for _id in _identities if _id is not None]
+
+        # noinspection PyBroadException
+        try:
+            notification = self.schedule.get('Notification')
+            notification_configurations_api = NotificationConfigurationsApi(status.session.client)
+            body = ReportNotificationConfigurationInputV1(
+                to_email_recipients=_lookup_identities(notification.get('To Email Recipients', [])),
+                cc_email_recipients=_lookup_identities(notification.get('CC Email Recipients', [])),
+                bcc_email_recipients=_lookup_identities(notification.get('BCC Email Recipients', [])),
+                report_format=notification.get('Report Format')
+            )
+            safely(
+                lambda: notification_configurations_api.set_notification_configuration_for_report(
+                    body=body, id=existing_annotation_id
+                ),
+                action_description=f'set notification for Annotation Schedule {existing_annotation_id}',
+                status=status
+            )  # NotificationConfigurationOutputV1
+
+        except NameError:
+            # We're on an earlier version of Seeq that doesn't have the NotificationConfigurationsApi
+            pass
+
     def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
         super().save(workbook_folder, include_rendered_content=include_rendered_content,
                      pretty_print_html=pretty_print_html)
