@@ -1,11 +1,13 @@
 from __future__ import annotations
 
+import datetime
 import types
 from typing import List, Optional
 
 import numpy as np
 import pandas as pd
 
+from seeq.base.seeq_names import SeeqNames
 from seeq.sdk import *
 from seeq.spy import _common
 from seeq.spy import _login
@@ -14,20 +16,24 @@ from seeq.spy._errors import *
 from seeq.spy._redaction import safely
 from seeq.spy._session import Session
 from seeq.spy._status import Status
+from seeq.spy._workbook_context import WorkbookContext
 from seeq.spy.workbooks import _folder
 from seeq.spy.workbooks import _pull
 from seeq.spy.workbooks import _user
 from seeq.spy.workbooks._annotation import Annotation
 from seeq.spy.workbooks._item import Item, ItemMap
 from seeq.spy.workbooks._template import ItemTemplate
-from seeq.spy.workbooks._workbook import Workbook
+from seeq.spy.workbooks._workbook import Workbook, WorkbookList, DatasourceMapList
 
 
+@Status.handle_keyboard_interrupt()
 def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datasource_map_folder=None,
          use_full_path=False, access_control=None, override_max_interp=False,
-         include_inventory=True, include_annotations: bool = True, scope_globals_to_workbook=False, refresh=True,
-         lookup_df: pd.DataFrame = None, specific_worksheet_ids: Optional[List[str]] = None, errors=None, quiet=None,
-         status: Optional[Status] = None, session: Optional[Session] = None):
+         include_inventory=None, include_annotations: bool = True, scope_globals_to_workbook=False, refresh=True,
+         lookup_df: pd.DataFrame = None, specific_worksheet_ids: Optional[List[str]] = None,
+         create_dummy_items_in_workbook=None, assume_dependencies_exist: bool = False,
+         errors=None, quiet=None, status: Optional[Status] = None,
+         session: Optional[Session] = None, item_map: Optional[ItemMap] = None) -> pd.DataFrame:
     """
     Pushes workbooks into Seeq using a list of Workbook object definitions.
 
@@ -41,10 +47,19 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         that a further subfolder hierarchy will be created to preserve the
         relative paths that the folders were in when they were searched for
         and pulled. If you specify None, then the workbook will stay where
-        it is (if it has already been pushed once). If you specify
-        spy.workbooks.MY_FOLDER, it will be moved to the user's home folder.
+        it is (if it has already been pushed once).
+
+        If you specify spy.workbooks.MY_FOLDER, it will be moved to the user's
+        home folder.
+
         If you specify a folder ID directly, it will be pushed to that folder.
 
+        If you specify spy.workbooks.ORIGINAL_FOLDER, it will be pushed to the
+        folder it was in when it was originally pulled. The folder hierarchy
+        will be recreated on the target server if it doesn't already exist.
+        (You must be an admin to use this to ensure you have permissions to
+        put things where they need to go.)
+
     owner : str, default None
         Determines the ownership of pushed workbooks and folders.
 
@@ -54,7 +69,8 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         All other options require that the logged-in user is an admin:
 
         If spy.workbooks.ORIGINAL_OWNER, ownership is assigned according to the
-        original owner of the pulled content.
+        original owner of the pulled content. (You must be an admin to use
+        this.)
 
         If spy.workbooks.FORCE_ME_AS_OWNER, existing content will be
         reassigned to the logged-in user.
@@ -117,10 +133,12 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         If True, then the Maximum Interpolation overrides from the source
         system will be written to the destination system.
 
-    include_inventory : bool, default True
+    include_inventory : bool, optional
         If True, then all calculated items that are scoped to the workbook
         will be pushed as well.
 
+        If omitted, SPy will push inventory in any non-template scenarios.
+
     include_annotations : bool, default True
         If True, downloads the HTML for Journal and Organizer Topic content.
 
@@ -150,6 +168,22 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         the push operation. No existing worksheets will be archived or moved
         in their ordering.
 
+    create_dummy_items_in_workbook : str, optional
+        If specified, then "dummy" items will be created for any stored items
+        that are not successfully mapped to the target system. This is useful
+        when you want to push a workbook to a system that doesn't have the same
+        datasources/items as the source system. The dummy items will be
+        created in the target system's "Seeq Data Lab" datasource and will
+        have the same name as the original item, and no data. They will be
+        scoped to the workbook specified.
+
+    assume_dependencies_exist : bool, default False
+        If True, then the push operation will assume that any dependencies
+        not found in the set of workbooks pushed are already present on the
+        target server. This is useful, for example, when you are pushing a
+        subset of workbooks (like a single Organizer Topic) and don't need
+        or want to push all of the referenced Workbench Analyses.
+
     errors : {'raise', 'catalog'}, default 'raise'
         If 'raise', any errors encountered will cause an exception. If
         'catalog', errors will be added to a 'Result' column in the status.df
@@ -159,7 +193,7 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
     quiet : bool, default False
         If True, suppresses progress output. Note that when status is
         provided, the quiet setting of the Status object that is passed
-        in takes precedent.
+        in takes precedence.
 
     status : spy.Status, optional
         If specified, the supplied Status object will be updated as the command
@@ -173,6 +207,9 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         store the login session state. This is useful to log in to different
         Seeq servers at the same time or with different credentials.
 
+    item_map:
+        For internal use only.
+
     Returns
     -------
     pandas.DataFrame
@@ -216,16 +253,36 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         (refresh, 'refresh', bool),
         (lookup_df, 'lookup_df', pd.DataFrame),
         (specific_worksheet_ids, 'specific_worksheet_ids', list),
+        (create_dummy_items_in_workbook, 'create_dummy_items_in_workbook', str),
+        (assume_dependencies_exist, 'assume_dependencies_exist', bool),
         (errors, 'errors', str),
         (quiet, 'quiet', bool),
         (status, 'status', Status),
-        (session, 'session', Session)
+        (session, 'session', Session),
+        (item_map, 'item_map', ItemMap)
     ])
 
-    status = Status.validate(status, quiet, errors)
     session = Session.validate(session)
+    status = Status.validate(status, session, quiet, errors)
     _login.validate_login(session, status)
 
+    if path == _folder.ORIGINAL_FOLDER and not use_full_path:
+        raise SPyValueError('You must specify use_full_path=True when path=spy.workbooks.ORIGINAL_FOLDER')
+
+    if path == _folder.ORIGINAL_FOLDER and not session.user.is_admin:
+        raise SPyValueError('Must be an admin to use path=spy.workbooks.ORIGINAL_FOLDER')
+
+    if owner == _user.ORIGINAL_OWNER and not session.user.is_admin:
+        raise SPyValueError('Must be an admin to use owner=spy.workbooks.ORIGINAL_OWNER')
+
+    if path == _folder.ORIGINAL_FOLDER and owner not in (_user.ORIGINAL_OWNER, _user.FORCE_ME_AS_OWNER):
+        raise SPyValueError(
+            'You must specify owner=spy.workbooks.ORIGINAL_OWNER or owner=spy.workbooks.FORCE_ME_AS_OWNER to use '
+            'path=spy.workbooks.ORIGINAL_FOLDER')
+
+    if label is not None and ('[' in label or ']' in label):
+        raise SPyValueError('label argument cannot contain square brackets []')
+
     owner_identity = None
     if owner not in [None, _user.ORIGINAL_OWNER, _user.FORCE_ME_AS_OWNER]:
         owner_identity = _login.find_user(session, owner)
@@ -233,7 +290,8 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
 
     status.update('Pushing workbooks', Status.RUNNING)
 
-    item_map = ItemMap(lookup_df=lookup_df)
+    if item_map is None:
+        item_map = ItemMap(lookup_df=lookup_df)
 
     if not isinstance(workbooks, list):
         workbooks = [workbooks]
@@ -241,10 +299,14 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
     # Make sure the datasource exists
     datasource_output = _metadata.create_datasource(session, datasource)
 
+    workbook_context: Optional[WorkbookContext] = None
+    if create_dummy_items_in_workbook is not None:
+        workbook_context = WorkbookContext.from_args(session, status, create_dummy_items_in_workbook, None, datasource)
+
     # Sort such that Analyses are pushed before Topics, since the latter usually depends on the former
     remaining_workbooks = list()
     sorted_workbooks = sorted(list(workbooks), key=lambda w: w['Workbook Type'])
-    status.df = pd.DataFrame(columns=['ID', 'Name', 'Type', 'Workbook Type', 'Count', 'Time', 'Result'])
+    status.df = pd.DataFrame(columns=['ID', 'Name', 'Type', 'Workbook Type', 'Count', 'Time', 'Errors', 'Result'])
     for index in range(len(sorted_workbooks)):  # type: int
         workbook = sorted_workbooks[index]
         remaining_workbooks.append((index, workbook))
@@ -253,17 +315,13 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         status.df.at[index, 'Type'] = workbook.type
         status.df.at[index, 'Workbook Type'] = workbook['Workbook Type']
         status.df.at[index, 'Count'] = 0
-        status.df.at[index, 'Time'] = 0
+        status.df.at[index, 'Time'] = datetime.timedelta(0)
+        status.df.at[index, 'Errors'] = 0
         status.df.at[index, 'Result'] = 'Queued'
 
-    datasource_maps = None
+    datasource_map_overrides: DatasourceMapList = DatasourceMapList()
     if datasource_map_folder:
-        datasource_maps = Workbook.load_datasource_maps(datasource_map_folder)
-        for datasource_map in datasource_maps:
-            # Specifying Override = True causes the StoredItem push code to try to look up an item based on the map
-            # instead of directly using the item's ID. This allows for pulling a workbook and then pushing it with a
-            # datasource map to swap the items within it.
-            datasource_map['Override'] = True
+        datasource_map_overrides = Workbook.load_datasource_maps(datasource_map_folder, overrides=True)
 
     folder_id = _create_folder_path_if_necessary(session, path, status)
     at_least_one_thing_pushed = False
@@ -277,6 +335,13 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
                 raise SPyRuntimeError('"workbooks" argument contains a non Workbook item: %s' % workbook)
 
             try:
+                status.reset_timer()
+
+                status.current_df_index = index
+                status.put('Count', 0)
+                status.put('Time', datetime.timedelta(0))
+                status.put('Result', 'Pushing')
+
                 if label is None:
                     # If a label is not supplied, check to see if we should be automatically isolating by user.
                     isolate_by_user = False
@@ -295,55 +360,73 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
                     if isolate_by_user:
                         label = owner_identity.username if owner_identity is not None else session.user.username
 
-                status.reset_timer()
-
-                status.current_df_index = index
-                status.put('Count', 0)
-                status.put('Time', 0)
-                status.put('Result', 'Pushing')
-
-                if datasource_maps is not None:
-                    workbook.datasource_maps = datasource_maps
-
-                workbook_folder_id = workbook.push_containing_folders(session, item_map, datasource_output,
-                                                                      use_full_path, folder_id, owner, label,
-                                                                      access_control, status)
+                if not isinstance(workbook, ItemTemplate):
+                    workbook.datasource_maps.extend(datasource_map_overrides.copy(), overwrite=True)
 
                 try:
-                    workbook.push(owner=owner, folder_id=workbook_folder_id, item_map=item_map, label=label,
-                                  datasource=datasource, access_control=access_control,
-                                  override_max_interp=override_max_interp, include_inventory=include_inventory,
-                                  include_annotations=include_annotations,
-                                  scope_globals_to_workbook=scope_globals_to_workbook,
-                                  specific_worksheet_ids=specific_worksheet_ids, status=status, session=session)
+                    include_inventory_for_this_workbook = (not isinstance(workbook, ItemTemplate)
+                                                           if include_inventory is None else include_inventory)
+
+                    workbook_folder_id = workbook.push_containing_folders(session, item_map, datasource_output,
+                                                                          use_full_path, folder_id, owner, label,
+                                                                          access_control, status)
+
+                    # Grab the success message now because already_pushed will always be true after the push
+                    success_message = 'Success'
+                    if workbook.already_pushed:
+                        success_message += ': Already pushed'
+
+                    try:
+                        workbook.push(owner=owner, folder_id=workbook_folder_id, item_map=item_map, label=label,
+                                      datasource=datasource, access_control=access_control,
+                                      override_max_interp=override_max_interp,
+                                      include_inventory=include_inventory_for_this_workbook,
+                                      include_annotations=include_annotations,
+                                      scope_globals_to_workbook=scope_globals_to_workbook,
+                                      specific_worksheet_ids=specific_worksheet_ids,
+                                      dummy_items_workbook_context=workbook_context,
+                                      status=status, session=session)
+                    except SPyException as e:
+                        status.raise_or_put(e, 'Result')
+                        continue
 
                     at_least_one_thing_pushed = True
                     at_least_one_thing_pushed_this_iteration = True
-                    remaining_workbooks.remove((index, workbook))
-                    status.put('Result', 'Success')
 
-                    if len(status.catalogued_errors) > 0:
-                        if status.errors == 'catalog':
-                            status.warn(status.catalogued_errors_str)
-                        else:
-                            raise SPyRuntimeError(status.catalogued_errors_str)
+                    remaining_workbooks.remove((index, workbook))
+
+                    status.put('Time', status.get_timer())
+                    status.put('Errors', len(workbook.push_errors))
+
+                    if len(workbook.push_errors) > 0:
+                        success_message += f', but with errors:\n{workbook.push_errors_str}'
+                        status.put('Result', success_message)
+                        if status.errors == 'raise':
+                            raise SPyRuntimeError(workbook.push_errors_str)
+                    else:
+                        status.put('Result', success_message)
 
                 except SPyDependencyNotFound as e:
                     status.put('Count', 0)
-                    status.put('Time', 0)
+                    status.put('Time', datetime.timedelta(0))
+                    status.put('Errors', 0)
                     status.put('Result', f'Need dependency: {str(e)}')
 
                     dependencies_not_found.append(str(e))
 
-            except BaseException as e:
-                _common.raise_or_catalog(status, e=e)
+            except ApiException as e:
+                raise SPyRuntimeError(_common.format_exception(e)) from e
 
         if not at_least_one_thing_pushed_this_iteration:
+            if assume_dependencies_exist and not item_map.fall_through:
+                item_map.fall_through = True
+                continue
+
             if status.errors == 'raise':
                 raise SPyRuntimeError('Could not find the following dependencies:\n%s\n'
                                       'Therefore, could not import the following workbooks:\n%s\n' %
                                       ('\n'.join(dependencies_not_found),
-                                       '\n'.join([str(workbook) for workbook in workbooks])))
+                                       '\n'.join([str(workbook) for _, workbook in remaining_workbooks])))
 
             break
 
@@ -352,24 +435,35 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
     new_workbooks = None
     if refresh and at_least_one_thing_pushed:
         refresh_workbook_inner_status = status.create_inner('Refresh Workbook')
-        pull_df = status.df.copy()
-        pull_df.drop(columns=['ID'], inplace=True)
-        pull_df.rename(columns={'Pushed Workbook ID': 'ID'}, inplace=True)
-        new_workbooks = _pull.pull(pull_df, include_inventory=include_inventory,
-                                   include_annotations=include_annotations, status=refresh_workbook_inner_status,
-                                   session=session)
+
+        new_workbooks = WorkbookList()
         for workbook in workbooks:
+            new_workbook_id = item_map[workbook.id]
+
+            specific_worksheet_ids_to_pull = None
+            if specific_worksheet_ids is not None:
+                specific_worksheet_ids_to_pull = [item_map[ws] for ws in specific_worksheet_ids]
+
+            include_inventory_for_refresh = include_inventory if include_inventory is not None else True
+            pulled_workbooks = _pull.pull(new_workbook_id, include_inventory=include_inventory_for_refresh,
+                                          specific_worksheet_ids=specific_worksheet_ids_to_pull,
+                                          include_annotations=include_annotations, status=refresh_workbook_inner_status,
+                                          session=session)
+
+            new_workbook = pulled_workbooks[0]
+            new_workbooks.append(new_workbook)
+
             if isinstance(workbook, ItemTemplate):
                 continue
 
-            new_workbook = [w for w in new_workbooks if w.id == item_map[workbook.id]][0]
-            workbook.refresh_from(new_workbook, item_map, include_inventory=include_inventory,
-                                  specific_worksheet_ids=specific_worksheet_ids, status=refresh_workbook_inner_status)
-            if folder_id is not None and folder_id != _common.PATH_ROOT:
+            workbook.refresh_from(new_workbook, item_map, include_inventory=include_inventory_for_refresh,
+                                  specific_worksheet_ids=specific_worksheet_ids, status=refresh_workbook_inner_status,
+                                  session=session)
+            if folder_id is not None and folder_id != _common.PATH_ROOT and folder_id != _folder.ORIGINAL_FOLDER:
                 workbook['Search Folder ID'] = folder_id
 
-    unique_results = status.df['Result'].drop_duplicates()
-    if len(unique_results) > 1 or (len(unique_results) == 1 and unique_results.iloc[0] != 'Success'):
+    max_errors = status.df['Errors'].max()
+    if max_errors > 0:
         status.update('Errors encountered, look at Result column in returned DataFrame', Status.FAILURE)
     else:
         status.update('Push successful', Status.SUCCESS)
@@ -383,7 +477,7 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
         item_map=item_map,
         status=status)
 
-    output_df = status.df
+    output_df = status.df.copy()
 
     _common.put_properties_on_df(output_df, output_df_properties)
 
@@ -391,6 +485,9 @@ def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datas
 
 
 def _create_folder_path_if_necessary(session: Session, path, status: Status):
+    if path == _folder.ORIGINAL_FOLDER:
+        return _folder.ORIGINAL_FOLDER
+
     if _common.is_guid(path):
         return path
 
@@ -428,20 +525,27 @@ def _create_folder_path_if_necessary(session: Session, path, status: Status):
             folder_filter = 'corporate'
             continue
 
+        kwargs = {
+            'filter': folder_filter,
+            'types': [SeeqNames.Types.folder],
+            'text_search': content_name,
+            'is_exact': True,
+            'limit': session.options.search_page_size,
+        }
         if parent_id:
-            folders = safely(lambda: folders_api.get_folders(filter=folder_filter,
-                                                             folder_id=parent_id,
-                                                             limit=10000),
-                             action_description=f'get Folders using filter "{folder_filter}" within {parent_id}',
-                             additional_errors=[400],
-                             status=status)  # type: WorkbenchItemOutputListV1
-        else:
-            folders = folders_api.get_folders(filter=folder_filter,
-                                              limit=10000)  # type: WorkbenchItemOutputListV1
+            kwargs['folder_id'] = parent_id
+        folders = safely(lambda: folders_api.get_folders(**kwargs),
+                         action_description=f'get Folders using filter "{folder_filter}" and name "{content_name}" '
+                                            f'within {parent_id}',
+                         additional_errors=[400],
+                         status=status)  # type: WorkbenchItemOutputListV1
 
         if folders is not None:
             for content in folders.content:  # type: WorkbenchSearchResultPreviewV1
                 if content.type == 'Folder' and content_name == content.name:
+                    if (parent_id is not None and content.ancestors is not None and len(content.ancestors) >= 1
+                            and content.ancestors[-1].id != parent_id):
+                        continue
                     existing_content_id = content.id
                     break
 
