@@ -343,7 +343,7 @@ def format_date_range_from_api_output(date_range_output: DateRangeOutputV1) -> d
             }
         # }
     # } catch (e) {
-    except BaseException as e:
+    except Exception as e:
         # logWarn(e);
         # formulaFormValid = false;
         formula_form_valid = False
