@@ -8,12 +8,11 @@ from typing import Optional, Union, Iterable
 
 import pandas as pd
 from deprecated import deprecated
-from mako.lookup import TemplateLookup
-from mako.template import Template
 
 from seeq import spy
+from seeq.base import util
 from seeq.sdk import *
-from seeq.spy import _common
+from seeq.spy import _common, _login
 from seeq.spy import _url
 from seeq.spy._common import safe_json_dumps
 from seeq.spy._errors import *
@@ -86,7 +85,7 @@ class Worksheet(Item):
     def pull(item_id, *, workbook=None, extra_workstep_tuples=None, include_annotations=True, include_images=True,
              session: Optional[Session] = None, status=None):
         session = Session.validate(session)
-        status = Status.validate(status)
+        status = Status.validate(status, session)
 
         if workbook is None:
             raise SPyValueError('workbook argument is None -- must be a valid Workbook object')
@@ -107,7 +106,7 @@ class Worksheet(Item):
 
     def pull_worksheet(self, session: Session, worksheet_id, extra_workstep_tuples=None, include_images=True,
                        status=None, include_annotations=True):
-        status = Status.validate(status)
+        status = Status.validate(status, session)
 
         if include_images and not include_annotations:
             raise SPyValueError('Must specify include_images=False if you specify include_annotations=False')
@@ -129,7 +128,7 @@ class Worksheet(Item):
         return f'Worksteps for {self.id}'
 
     def push(self, session: Session, pushed_workbook_id, item_map, datasource_output, existing_worksheet_identifiers,
-             include_inventory, include_annotations, label=None, status: Status = None):
+             include_inventory, include_annotations, access_control, label=None, status: Status = None):
         self._item_map_for_push = item_map
 
         existing_worksheet_id = None
@@ -174,14 +173,18 @@ class Worksheet(Item):
 
     @property
     def referenced_items(self):
-        referenced_items = list()
+        referenced_items = set()
         if hasattr(self, 'worksteps'):
             for workstep_id, workstep in self.worksteps.items():  # type: Workstep
-                referenced_items.extend(workstep.referenced_items)
+                referenced_items.update(workstep.referenced_items)
 
-        referenced_items.extend(self._annotation.referenced_items)
+        referenced_items.update(self._annotation.referenced_items)
 
-        return referenced_items
+        return list(referenced_items)
+
+    def referenced_items_df(self):
+        unique_ids = {r.id for r in self.referenced_items}
+        return pd.DataFrame([self.workbook.item_inventory[_id].definition_dict for _id in unique_ids])
 
     @property
     def referenced_worksteps(self):
@@ -211,7 +214,7 @@ class Worksheet(Item):
     def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
         worksheet_json_file = Worksheet._get_worksheet_json_file(workbook_folder, self.id)
 
-        with open(worksheet_json_file, 'w', encoding='utf-8') as f:
+        with util.safe_open(worksheet_json_file, 'w', encoding='utf-8') as f:
             json.dump(self._definition, f, indent=4)
 
         self._annotation.save(workbook_folder, include_rendered_content=include_rendered_content,
@@ -220,7 +223,7 @@ class Worksheet(Item):
     def _load(self, workbook_folder, worksheet_id):
         worksheet_json_file = Worksheet._get_worksheet_json_file(workbook_folder, worksheet_id)
 
-        with open(worksheet_json_file, 'r', encoding='utf-8') as f:
+        with util.safe_open(worksheet_json_file, 'r', encoding='utf-8') as f:
             self._definition = json.load(f)
 
         if self.workbook['Workbook Type'] == 'Analysis':
@@ -229,7 +232,8 @@ class Worksheet(Item):
             self._annotation = Report.load(self, workbook_folder)
 
         self.worksteps = dict()
-        workstep_files = glob.glob(os.path.join(workbook_folder, 'Worksheet_%s_Workstep_*.json' % worksheet_id))
+        folder_escaped = glob.escape(workbook_folder)
+        workstep_files = util.safe_glob(os.path.join(folder_escaped, 'Worksheet_%s_Workstep_*.json' % worksheet_id))
         for workstep_file in workstep_files:
             match = re.match(r'.*?Worksheet_.*?_Workstep_(.*?).json$', workstep_file)
             workstep_id = match.group(1)
@@ -418,8 +422,8 @@ class AnalysisWorksheet(Worksheet):
         >>> worksheet = workbook.worksheets['My Worksheet']
         >>> worksheet.pull_current_workstep()
         """
-        status = Status.validate(status, quiet)
         session = Session.validate(session)
+        status = Status.validate(status, session, quiet)
         self.pull_worksheet(session, self.id, include_images=False, status=status, include_annotations=False,
                             include_referenced_worksteps=False)
 
@@ -454,8 +458,8 @@ class AnalysisWorksheet(Worksheet):
         >>>     worksheet.display_items[worksheet.display_items['Name'] == 'Temperature'].index)
         >>> worksheet.push_current_workstep()
         """
-        status = Status.validate(status, quiet)
         session = Session.validate(session)
+        status = Status.validate(status, session, quiet)
         workstep = self.current_workstep()
         workstep_input = WorkstepInputV1(data=safe_json_dumps(workstep.data))
         safely(lambda: WorkbooksApi(session.client).create_workstep(workbook_id=self.workbook.id, worksheet_id=self.id,
@@ -463,11 +467,11 @@ class AnalysisWorksheet(Worksheet):
                status=status)
 
     def push(self, session: Session, pushed_workbook_id, item_map, datasource_output, existing_worksheet_identifiers,
-             include_inventory, include_annotations, label=None, status: Status = None):
-        status = Status.validate(status)
+             include_inventory, include_annotations, access_control, label=None, status: Status = None):
+        status = Status.validate(status, session)
         worksheet_output = super().push(session, pushed_workbook_id, item_map, datasource_output,
                                         existing_worksheet_identifiers, include_inventory,
-                                        include_annotations, label=label, status=status)
+                                        include_annotations, access_control, label=label, status=status)
 
         existing_workstep_walk = ExistingWorkstepWalk(pushed_workbook_id, worksheet_output)
 
@@ -504,14 +508,14 @@ class AnalysisWorksheet(Worksheet):
 
         if include_annotations:
             self.journal.push(session, pushed_workbook_id, worksheet_output.id, item_map, datasource_output,
-                              push_images=True, label=label)
+                              access_control, push_images=True, label=label)
 
         return worksheet_output
 
-    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
-        super().refresh_from(new_item, item_map, status)
+    def refresh_from(self, new_item, item_map: ItemMap, status: Status, session: Session):
+        super().refresh_from(new_item, item_map, status, session)
 
-        self.journal.refresh_from(new_item.journal, item_map, status)
+        self.journal.refresh_from(new_item.journal, item_map, status, session)
 
         for workstep_id, workstep in self.worksteps.items():
             workstep['ID'] = item_map[workstep_id]
@@ -548,6 +552,8 @@ class AnalysisWorksheet(Worksheet):
             ================= =================================== =============
             Color             A 3-part RGB hexadecimal color spec
                               starting with "#".  E.g. #CE561B    All
+            Axis Show         Boolean if the axis should be
+                              shown                               Si, Sc, M
             Axis Auto Scale   Boolean if the axis should auto
                               scale                               Si, Sc, M
             Axis Limits       A dict of {'min': float,
@@ -557,19 +563,21 @@ class AnalysisWorksheet(Worksheet):
             Axis Group        An identifier to specify shared
                               axes                                Si, Sc, M
             Lane              The lane a trend is plotted in      Si, Sc, M
-            Align             Specify the side of the plot for
+            Axis Align        Specify the side of the plot for
                               the y-axis. 'Left' (default) or
                               'Right'.                            Si, Sc, M
-            Line              The trend line style. Options are
+            Line Style        The trend line style. Options are
                               'Solid', 'Dot', 'Short Dash',
                               'Long Dash', 'Short Dash-Dot',
                               'Short Dash-Dot-Dot',
                               'Long Dash-Dot',
                               'Long Dash-Dot-Dot', 'Dash-Dot'     Si, Sc, M
             Line Width        The width of the line.              Si, Sc, M
-            Samples           The sample display style. Options
+            Samples Display   The sample display style. Options
                               are 'Line', 'Line and Sample',
                               'Samples', 'Bars'                   Si
+            Selected          Boolean indicating if item should
+                              be selected                         All
             Stack             Boolean indicating if bars should
                               be stacked                          T
             ================= =================================== =============
@@ -916,29 +924,35 @@ class TopicDocument(Worksheet):
         return self._annotation
 
     def pull_rendered_content(self, session: Session, status: Status):
-        status = Status.validate(status)
+        status = Status.validate(status, session)
 
         if self.report:
             self.report.pull_rendered_content(session, status)
 
     def push(self, session: Session, pushed_workbook_id, item_map, datasource_output, existing_worksheet_identifiers,
-             include_inventory, include_annotations, label=None, status: Status = None):
+             include_inventory, include_annotations, access_control, label=None, status: Status = None):
         worksheet_output = super().push(session, pushed_workbook_id, item_map, datasource_output,
                                         existing_worksheet_identifiers, include_inventory, include_annotations,
-                                        label=label, status=status)
+                                        access_control, label=label, status=status)
 
         if include_annotations:
             self.report.push(session, pushed_workbook_id, worksheet_output.id, item_map, datasource_output,
-                             push_images=True, label=label)
+                             access_control, push_images=True, label=label, status=status)
 
         return worksheet_output
 
-    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
-        super().refresh_from(new_item, item_map, status)
+    def refresh_from(self, new_item, item_map: ItemMap, status: Status, session: Session):
+        super().refresh_from(new_item, item_map, status, session)
 
-        self.report.refresh_from(new_item.report, item_map, status)
+        self.report.refresh_from(new_item.report, item_map, status, session)
 
     def render_template(self, text=None, filename=None, **kwargs):
+        try:
+            from mako.lookup import TemplateLookup
+            from mako.template import Template
+        except ImportError:
+            raise SPyDependencyNotFound('`Mako` is required to render templates. Please install it using '
+                                        f'`pip install seeq-spy[templates]` to use this feature.')
         if filename:
             sibling_folder_lookup = TemplateLookup(directories=[os.path.dirname(filename)])
             template = Template(filename=filename, lookup=sibling_folder_lookup, input_encoding='utf-8')
@@ -1046,7 +1060,15 @@ class TopicDocument(Worksheet):
         self.report.schedule = value
 
 
-def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, quiet: Optional[bool] = False):
+@deprecated(reason='Use spy.utils.pull_worksheet_via_url instead')
+def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, quiet: Optional[bool] = False,
+                                    session: Optional[Session] = None):
+    return pull_worksheet_via_url(url=url, include_archived=include_archived, quiet=quiet, session=session,
+                                  minimal=False)
+
+
+def pull_worksheet_via_url(url: str, *, minimal: bool = True, include_archived: bool = False,
+                           quiet: Optional[bool] = False, session: Optional[Session] = None):
     """
     Get a worksheet from a valid Workbench Analysis URL.
 
@@ -1054,11 +1076,17 @@ def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, qu
     ----------
     url : str
         The URL of the Workbench Analysis
+    minimal : bool, default True
+        If True, only the minimal amount of information will be pulled from the server. This is useful if you are
+        only interested in, for example, knowing what display_items are present.
     include_archived : bool, default False
         If True, returns the worksheet even if the workbook is archived. If False, raises an exception
         if the workbook is archived.
     quiet : bool, default False
         If True, suppresses progress output.
+    session : spy.Session, optional
+        If supplied, the Session object used for the connection to Seeq
+        Server. If not supplied, the default session is used.
 
     Returns
     -------
@@ -1072,17 +1100,23 @@ def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, qu
         raise SPyValueError(f'The supplied URL is not a valid Seeq address. Verify that both the workbook ID and '
                             f'worksheet ID are valid Seeq references')
 
-    workbook_search = spy.workbooks.search({'ID': workbook_id}, status=Status(errors='catalog', quiet=quiet))
+    session = Session.validate(session)
+    status = Status(errors='catalog', quiet=quiet)
+    _login.validate_login(session, status)
+
+    workbook_search = spy.workbooks.search({'ID': workbook_id}, status=status, session=session)
     if len(workbook_search) == 0:
         raise SPyRuntimeError(f'Could not find workbook with ID "{workbook_id}"')
     workbooks = spy.workbooks.pull(workbook_search, include_referenced_workbooks=False, include_inventory=False,
-                                   specific_worksheet_ids=[worksheet_id], status=Status(quiet=quiet))
+                                   include_annotations=not minimal, include_images=not minimal,
+                                   include_access_control=not minimal, specific_worksheet_ids=[worksheet_id],
+                                   status=Status(quiet=quiet), session=session)
     if not workbooks:
         raise SPyValueError(f"An error occurred while pulling Worksheet '{worksheet_id}' from Workbook '{workbook_id}'")
     workbook = workbooks[0]
 
     if _common.get(workbook.definition, 'Archived') and not include_archived:
-        raise SPyValueError(f"Workbook '{workbook_id}' is archived. Supply 'include_archive=True' if"
+        raise SPyValueError(f"Workbook '{workbook_id}' is archived. Supply 'include_archived=True' if "
                             f"you want to retrieve the items of an archived workbook")
 
     if workbook.definition['Workbook Type'] != 'Analysis':
@@ -1092,7 +1126,7 @@ def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, qu
     try:
         worksheet = [x for x in workbook.worksheets if x.id == worksheet_id][0]
         if _common.get(worksheet.definition, 'Archived') and not include_archived:
-            raise SPyValueError(f"Worksheet '{worksheet_id}' is archived. Supply 'include_archive=True' if"
+            raise SPyValueError(f"Worksheet '{worksheet_id}' is archived. Supply 'include_archived=True' if "
                                 f"you want to retrieve archived items")
         return worksheet
     except IndexError:
